/***************************************************************************************/
/**                              CAN Sim DLL API
 *                       (c)2019-2020 Digiteq automotive
 ****************************************************************************************/
#ifndef _CANSIMAPI_H_
#define _CANSIMAPI_H_


// Common includes
#include <stdint.h>
#include <stdio.h>
#include <stdbool.h>

// Lib exporting macro
#ifndef DllExport
# if defined(_WIN32)
 #if defined BUILD_CANSIM_DLL
  #  define DllExport __declspec(dllexport)
 #else
  #ifdef BUILD_CANSIM_STATIC
    #define DllExport
  #else
    #define DllExport __declspec(dllimport)
  #endif
 #endif
# else
#  define DllExport __attribute__ ((visibility("default")))
# endif
#endif  // DllExport


#ifndef __cplusplus
typedef void *CansimHandler;
#endif  // !__cplusplus

#ifdef __cplusplus
 #ifdef _CANSIMAPI_INT_H_
 class Cansim;
 typedef Cansim* CansimHandler;
 #else
 typedef void *CansimHandler;	/**< Customer access from C++ */
 #endif
extern "C" {
#endif


typedef void(*PrintFunc)(const char*);

typedef enum LOG_LEVELS_CANSIM {
    ERROR_LEVEL_CANSIM,		/**< Error messages */
    INFO_LEVEL_CANSIM,		/**< Informative/non-critical messages */
    COMM_LEVEL_CANSIM,		/**< Serial communication */
} LOG_LEVELS_CANSIM;


/** @struct tagCanMessage
 * @brief This structure contains all possible options for the
 * configuration of a CAN message for transmitting by Cansim using 'Cansim::AddMessage'.
 * Provides an alternative option to call the method 'Cansim::AddMessage'
 * with all these arguments, but does same thing.
 * After the method call the structure may be destroyed.
 * @var tagCanMessage::CanMap
 * @brief Member 'CanMap' contains **mask** of output CAN interfaces
 * 1. bit - Infotaiment
 * 2. bit - Konfort
 * 3. bit - Abtrieb
 * 4. bit - Fahrwerk
 * @var tagCanMessage::CanGatewayNumber
 * @brief Member 'CanGatewayNumber' should contains same value as 'CanMap' (not implemented).
 * @var tagCanMessage::ID
 * @brief Member 'ID' contains CAN message ID (length depends if is Extended ID enabled).
 * @var tagCanMessage::ExtendedID
 * @brief Member 'ExtendedID' contains flag if message use extended CAN ID.
 * @var tagCanMessage::FastSlowSend
 * @brief Message is transmitted with tagCanMessage::PeriodFast
 * @var tagCanMessage::FD
 * @brief Member 'FD' flag if is CAN FD enabled
 * @var tagCanMessage::Error
 * @brief TODO(JKI): Nevim je potreba dodelat - souvisi s CAN FD.
 * @var tagCanMessage::BRS
 * @brief Member 'FD' flag if is CAN FD BRS enabled.
 * @var tagCanMessage::GroupID
 * @brief Member 'GroupID' contains an ID of group for bulk management
 * @var tagCanMessage::Enable
 * @brief Member 'Enable' contains a flag if the message will be enabled instantly
 * after Cansim receive configuration command.
 * @var tagCanMessage::DLC
 * @brief Member 'DLC' contains a CAN Data Length Code.
 * https://elearning.vector.com/mod/page/view.php?id=368
 * @var tagCanMessage::InhibitTime
 * @brief Member 'InhibitTime' contains inhibit time of message in milliseconds.
 * The definition of inhibit time is in CAN WV Lastenheft.
 * @var tagCanMessage::Period
 * @brief Member 'Period' contains periode time of the CAN message in milliseconds
 * @var tagCanMessage::PeriodFast
 * @brief Member 'Period' contains fast sending period in milliseconds
 * The definition of inhibit time is in CAN WV Lastenheft. */
typedef struct tagCanMessage {
    uint8_t DLC;  // Data length
    uint8_t CanMap;  //
    uint8_t CanGatewayNumber;  // Kde je gateway
    uint32_t ID;
    bool ExtendedID;
    bool FD;  // FD {CAN with Flexible Data-Rate}
    bool Error;  // FD
    bool BRS;  // FD {Bit Rate Switch}
    bool GroupID;
    bool Enable;
    bool FastSlowSend;
    uint16_t InhibitTime;
    uint16_t Period;
    uint16_t PeriodFast;
} CanMessage, *PtrCanMessage;

typedef struct tagRecvMsgCansim {
    int32_t Handler;
    uint16_t CanId16;
    uint8_t CanPort;
    uint8_t DLC;
    bool ExtendedID;
    bool FD;
    bool Error;
    bool BRS;
    uint32_t timestamp;
} RecvMsgCansim, *PtrRecvMsgCansim;


/**
 * @brief This method converts DLC to length according to: https://elearning.vector.com/mod/page/view.php?id=368.
 *
 * @param DLC - Data Length Code. If the value exceed 0x0f than is trimmed by 0x0f mask.
 * @return **uint8_t** - Length in bits.
 */
DllExport uint8_t DLC2length(uint8_t* DLC);

/**
 * @brief
 *
 * @return CansimHandler
 */
DllExport CansimHandler NewCansim(void);

/**
 * @brief
 *
 * @param Handle
 * @return DllExport deleteCansim
 */
DllExport void DeleteCansim(CansimHandler Handle);

DllExport void DisconnectCansim(CansimHandler Handle);

DllExport int32_t StartReceiverCansim(CansimHandler Handle);
DllExport void StopReceiverCansim(CansimHandler Handle);

DllExport void DebugCommCansim(CansimHandler Handle, const char *FileName);

DllExport int32_t RebootCansim(CansimHandler Handle, const bool Jump);
DllExport int32_t ResetCansim(CansimHandler Handle);
DllExport int32_t BootloaderCansim(CansimHandler Handle);
DllExport int32_t GetVersionCansim(CansimHandler Handle);

DllExport int32_t ConnectCansim(CansimHandler Handle,
                                const char* COMPort,
                                const uint32_t ReadTimeout,
                                const uint32_t WriteTimeout);
DllExport int32_t AddMessageCansim(CansimHandler Handle, const PtrCanMessage msg,
                                   const uint8_t* Data);
DllExport int32_t RemoveMessageCansim(CansimHandler Handle, const uint16_t MsgHandler);
DllExport int32_t EnableMessageCansim(CansimHandler Handle, const uint16_t MsgHandler);
DllExport int32_t DisableMessageCansim(CansimHandler Handle, const uint16_t MsgHandler);
DllExport int32_t SetMessageDataCansim(CansimHandler Handle, const uint16_t MsgHandler,
                                       const uint8_t* Data, const uint8_t* DataMask,
                                       const bool Async, const bool response);
DllExport int32_t GetHandlerCansim(CansimHandler Handle, const uint8_t CanMap,
                                   const uint8_t CanGatewayNumber, const uint32_t CanID,
                                   const bool ExtendedCanID, const bool FastSlowSend, const bool FD);
DllExport int32_t EnableRequestCansim(CansimHandler Handle, const uint16_t ReqHandler);
DllExport int32_t DisableRequestCansim(CansimHandler Handle, const uint16_t ReqHandler);
DllExport int32_t RemoveRequestCansim(CansimHandler Handle, const uint16_t ReqHandler);
DllExport int32_t RequestMessageCansim(CansimHandler Handle,
                                       const uint8_t CanMap,
                                       const uint32_t CanID,
                                       const bool Enabled,
                                       const bool ExtendedCanID,
                                       const uint8_t DLC,
                                       const bool CheckDLC,
                                       const bool ReturnID,
                                       const uint8_t* Mask);
DllExport int32_t RequestSignalCansim(CansimHandler Handle,
                                      const int32_t MsgHandler,
                                      const uint16_t StartSymbol,
                                      const uint16_t StopSymbol);
/**
 * @brief This method reads cached 32 bit value of the signal registered by Cansim::RequestSignal.
 *
 * @param SigHandler - Handler of the requested message.
 * @param Value - Pointer to allocated space for return value.
 * If is nullptr, than command is nonsense.
 * @param TimeFromReceive - Get time since message has been read from serial. Milliseconds. \n
 * If is nullptr, than it save some time.
 * @param GetLatest - NOT IMPLEMENTED. \n
 * In future: If true than is returned last received signal else next message in buffer (queue).
 * @return **int32_t**\n
 * If the return value bigger than zero so it is signal value. It is maximal 16 bit unsigned integer.\n
 * If the response is smaller than zero - some error ocured:\n
 * **-1** - Any cached signal value yet.\n
 * **-2** - invalid request signal call.\n
 * **-3** - Thread does not run.\n
 * **-1000** - NOT IMPLEMENTED.\n
 */
DllExport int32_t SignalReadValue32Cansim(CansimHandler Handle,
                                        const uint16_t SigHandler,
                                        uint32_t *Value,
                                        uint32_t *TimeFromReceive,
                                        bool GetLatest);
/**
 * @brief This method reads cached 64 bit value of the signal registered by Cansim::RequestSignal.
 *
 * @param SigHandler - Handler of the requested message.
 * @param Value - Pointer to allocated space for return value.
 * If is nullptr, than command is nonsense.
 * @param TimeFromReceive - Get time since message has been read from serial. Milliseconds. \n
 * If is nullptr, than it save some time.
 * @param GetLatest - NOT IMPLEMENTED. \n
 * In future: If true than is returned last received signal else next message in buffer (queue).
 * @return **int32_t**\n
 * If the return value bigger than zero so it is signal value. It is maximal 16 bit unsigned integer.\n
 * If the response is smaller than zero - some error ocured:\n
 * **-1** - Any cached signal value yet.\n
 * **-2** - invalid request signal call.\n
 * **-3** - Thread does not run.\n
 * **-1000** - NOT IMPLEMENTED.\n
 */
DllExport int32_t SignalReadValue64Cansim(CansimHandler Handle,
                                        const uint16_t SigHandler,
                                        uint64_t *Value,
                                        uint32_t *TimeFromReceive,
                                        bool GetLatest);

DllExport int32_t AddSignalCansim(CansimHandler Handle,
                                      const uint16_t MsgHandler,
                                      const uint8_t Group,
                                      const uint16_t StartSymbol,
                                      const uint16_t StopSymbol,
                                      const uint8_t MuxInfo);

DllExport int32_t SignalChangeValueCansim(CansimHandler Handle,
                                          const uint16_t SigHandler,
                                          const uint16_t Value,
                                          const bool async,
                                          const bool response);
DllExport int32_t BlobRemoveCansim(CansimHandler Handle);

DllExport int32_t RemoveSignalCansim(CansimHandler Handle, const uint16_t SigHandler);
DllExport int32_t ExecuteSignalChangeCansim(CansimHandler Handle, const bool response);
DllExport int32_t SetSignalControlCansim(CansimHandler Handle,
                                         const uint16_t SigHandler,
                                         const uint8_t ControlNumber,
                                         const uint32_t MinValue,
                                         const uint32_t MidValue,
                                         const uint32_t MaxValue);
DllExport int32_t ChangeSignalSgroupCansim(CansimHandler Handle,
                                           const uint8_t Sgroup,
                                           const uint16_t Value);

DllExport int32_t AddCrcCansim(CansimHandler Handle, const uint8_t* S_PDU_Kennungsfolge);
DllExport int32_t ChangeCrcCansim(CansimHandler Handle, const uint16_t MsgHandler, uint8_t* S_PDU_Kennungsfolge);
DllExport int32_t DisableCrcCansim(CansimHandler Handle, const uint16_t MsgHandler);

DllExport int32_t GetInternalThreadStatus(CansimHandler Handle);

/**
 * @brief Read buffered asynchronous message - nonblocking.
 *
 * @param Handle - handler to the Cansim.
 * @param MsgHandler - Handler of message registered by Cansim::RequestMessage
 * @param DataStruct - Structure for asynchronous messages. \n
 * @param Data - Pointer to allocated space for CAN messages data. \n
 * NOTE: If is member **Data** = NULL, than data are not copied (could save some time).
 * @param Length - Length of the allocated buffer. If is smaller than CAN message
 * payload length, than are copied only lower bytes \n
 * @return **int32_t** \n
 * If the function succeeds, the return value is bigger or equal to **zero**.
 * And returns number of bytes written to **Data**.
 * Value less 0 could means that message has not been received yet.
 */
DllExport int32_t ReadAsyncMsgCansim(CansimHandler Handle,
                                     const int32_t MsgHandler,
                                     PtrRecvMsgCansim DataStruct,
                                     uint8_t* Data,
                                     const uint8_t Length);

/**
 * @brief Read Cansim internal EEPROM memory
 *
 * @param Handle - Cansim handler
 * @param address - Start address to read
 * @param length - Data length in bytes
 * @param data - Output data buffer, must be >= length
 * @return int32_t - 0 success
 */
DllExport int32_t ReadEepromCansim(CansimHandler Handle,
                                   const uint16_t address,
                                   const uint8_t length,
                                   uint8_t* data);

/**
 * @brief Write to internal EEPROM memory
 *
 * @param Handle - Cansim handler
 * @param address - Start address to write
 * @param data - Data to write
 * @param length - Data length
 * @return int32_t - 0 success
 */
DllExport int32_t WriteEepromCansim(CansimHandler Handle,
                                    const uint16_t address,
                                    const uint8_t length,
                                    const uint8_t* data);

/**
 * @brief Get the Cansim unix time.
 *
 * @param Handle - Cansim handler
 * @param time - time from cansim - internally max size 32bit.
 * @return int32_t - 0 success
 */
DllExport int32_t GetTimeCansim(CansimHandler Handle, time_t* time);

/**
 * @brief Set the Cansim unix time.
 *
 * @param Handle - Cansim handler
 * @param time - UNIX time to write - internally max size 32bit.
 * @return int32_t - 0 success
 */
DllExport int32_t SetTimeCansim(CansimHandler Handle, time_t time);

/* Printer wrapper */
DllExport int32_t SetPrintFileCansim(CansimHandler Handle, FILE* PrintStream, LOG_LEVELS_CANSIM Level);
DllExport int32_t ClearPrintFileCansim(CansimHandler Handle, FILE* PrintStream, LOG_LEVELS_CANSIM Level);
DllExport int32_t SetPrintFunctionCansim(CansimHandler Handle, PrintFunc, LOG_LEVELS_CANSIM Level);
DllExport int32_t ClearPrintFunctionCansim(CansimHandler Handle, PrintFunc, LOG_LEVELS_CANSIM Level);

DllExport void PrintfErrCansim(CansimHandler Handle, const char* Text, ...);
DllExport void PrintfInfCansim(CansimHandler Handle, const char* Text, ...);

#ifdef __cplusplus
}
#endif

#endif  // CANSIMAPI_CANSIMAPI_H_
